// Programme pour wagon odometre,tachymetre 
//
// V1.0 	S.G.	création du logiciel
// 		Calcul distance, vitesse
// 		Affichage OLED, serial BlueTooth
//        r reset distance et vitesse
//        a allume led feux arrieres
//        e eteint led feux arrieres
//        u changement d'unite pour l'affichage
// 		Clignotement led temoin de presence d'IT
//
#include <FlexiTimer2.h>      // configuration automatisee des IT du timer 2
                
// Declarations pour affichage OLED SSD1306
#include <SPI.h>              // utilisee par la bibliotheque Adafruit pilotant l'afficheur
#include <Wire.h>             // bibliotheque I2C
#include <Adafruit_GFX.h>     // bibliotheque graphique Adafruit
#include <Adafruit_SSD1306.h> // bibliotheque afficheur
// Declarations limitees car l'afficheur OLED SSD1306 est connecte en I2C (SDA, SCL pins)
#define SCREEN_WIDTH 128      // OLED largeur de l'afficheur en pixels
#define SCREEN_HEIGHT 64      // OLED hauteur de l'afficheur en pixels
#define OLED_RESET     -1     // n° broche relié au reset de l'afficheur (-1 si reset commun avec l'arduino)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET); // instanciation de l'afficheur

#define PIN_X 0                         // analog pin pour mesurer x (0 à 1023 pour 0 à 5V)
#define PIN_Y 1                         // analog pin pour mesurer y (0 à 1023 pour 0 à 5V)
#define X_OFFSET 512                    // valeur du 0 pour le capteur x  (2,5 V)
#define Y_OFFSET 512                    // valeur du 0 pour le capteur y  (2,5 V)

const double _2PI = 2 * 3.1415926535;   // 2 Pi pour les calculs en radians
const double CSTE_ROUE = (11./4.);      // (mm/rad) diamètre roue/2 -> rayon  puis /2 car 2 paires d'aimants par tour de roue
const double _echantil = 0.002;         // periode d'echantillonnage pour le filtre
const double _Wn = 2.*3.1415926525 * 3.;  // rad/s   ici Fc = 3Hz pour le filtre de calcul de la vitesse
const double _2Ksi = 2. * 1.;           // ksi =1  (filtre super amorti).
const double COEF_DIST=87./1000.;       // passage de mm à km avec echelle HO
const double COEF_VIT=3.6/1000.*87.;    // passage de mm/s à km/h avec echelle HO

// les variables suivantes sont globales car l'interruption en a besoin a chaque calcul
double X1 = 0, X2 = 0;                  // variables d'etat pour le filtre 2eme ordre de calcul de la vitesse
bool init_dist = true;                  // remet à 0 la distance affichée quand vrai
double offset_dist = 0.;                // decalage de zero appliqué a la distance affichée
double coef_dist = 1.;                  // coef pour le changement d'unite de distance
double coef_vit = 1.;                   // coef pour le changement d'unite de vitesse
float angle_1 = 0;                      // sauvegarde du calcul d'angle precedent pour detection des passages +/- pi
float tours = 0;                        // sauvegarde du nombre de tours deja faits pour calcul de la distance totale
double distance;                        // distance parcourue (accessible partout)
float vitesse;                          // vitesse apres filtrage (accessible partout)
int compteur=0;                         // compte les IT, utilisé pour faire clignoter la led témoin de "vie" du soft et des IT
                                        // repasse à zero tout seul au bout de 65536 IT (131s)
										
// variables globales pour générer les trames à envoyer
char  dist[11]="        mm";            // chaines utilisées pour formater l'affichage
char   vit[11]="      mm/s";            // 10 caracteres affiches en taille 2 + le caractere de fin '\0'
int  unit_mm = 0;                       // memorise la selection du type d'unites

// donnees globales pour la gestion de la periode d'envoi
long aff_date;                          // date du dernier raffraichissement de l'ecran OLED
#define AFF_UPDATE  100                 // periode de rafraichissement de l'affichage en ms

#define led_pin 13                      // led à faire clignoter (temoin de "vie" des IT
#define lanterne_pin 5                  // feux arrière du wagon
#define test_IT_pin 6                   // sortie digitale pour chronometrer les IT avec un oscilloscope
#define test_Form_pin 7                 // sortie digitale pour chronometrer le formatage
#define test_Com_pin 8                  // sortie digitale pour chronometrer l'affichage
#define test_Aff_pin 9                  // sortie digitale pour chronometrer la liaison PC

// variables intermédiaires du calcul declarees globales pour pouvoir les utiliser en debug via la liaison série
int x, y;                               // mesures lues (cos et sin)
double angle;                           // angle magnetique de l'essieu (2 paires de poles)
double dx1_dt, dx2_dt;                  // variables intermédiaires pour le 2eme ordre de calcul de la vitesse

void interruption()  {
    digitalWrite(test_IT_pin, HIGH);              // positionnement sortie de chronometrage interruption
// mesure de l'angle "magnetique" = angle_essieu * 2 (car 2 paires de poles)
    x=analogRead(PIN_X)-X_OFFSET;                 // mesure cos et sin de l'angle magnetique
    y=analogRead(PIN_Y)-Y_OFFSET;                 // peut importe le gradiant de mesure, tant qu'il est identique sur X et Y
    angle = atan2(y, x);                          // extraction de l'angle (calcul ratiometrique)
// decompte du nombre de tours
    if ((angle - angle_1) > 4 ) tours -= _2PI;    // l'angle vient de faire un saut a -Pi
    if ((angle - angle_1) < -4 ) tours += _2PI;   // l'angle vient de faire un saut a +Pi
    angle_1 = angle;                              // memorisation de l'angle pour IT suivante
    distance = (angle + tours) * CSTE_ROUE;       // distance = angle * Rayon / nbre paires de poles
    if (init_dist) {                              // raz distance
      X1 = 0; X2 = 0;                             // raz variables d'etat du filtre
      tours=0;                                    // raz nbre de tours
      offset_dist = angle_1*CSTE_ROUE;            // reglage de l'offset pour avoir distance = 0
      init_dist = false;                          // raz flag pour ne pas repasser ici
    }
    distance -= offset_dist;                      // compensation de la position initiale de l'essieu
    
// filtre derivee (2eme ordre)
    dx1_dt = (distance - X2 - _2Ksi * X1) * _Wn;  // equation d'evolution de l'etat X1
    dx2_dt = X1 * _Wn;                            // equation d'evolution de l'etat X2
    X1 += dx1_dt * _echantil;                     // integration X1
    X2 += dx2_dt * _echantil;                     // integration X2
    vitesse = dx2_dt;                             // X2 recopie la distance à travers le filtre
                                                  //     ==> sa derivee est la vitesse filtree
// pour faire clignoter la led temoin de "vie" du systeme                                                                                                                  
    digitalWrite(led_pin, !(compteur++ & 0x100)); // le bit 2^9 du compteur change toutes les 256 IT 
                                                  // soit 256*2ms = 512ms la led clignote a 1,024 Hz
    digitalWrite(test_IT_pin, LOW);               // positionnement sortie de chronometrage de l'interruption
}


void serialEvent() {                    // fonction appelee par le systeme si il y a quelquechose de recu sur
                                        // la liaison serie (ici on ne fait que quelques commutations de flags).
  char c;                               // caractere intermediaire recu
  
  if (Serial.available()>0)             // y a-t-il qquechose recue venant de la liaison serie
        c=Serial.read();                // lecture du 1er caractere du buffer de reception
  switch(c) {                           // aiguillage en fonction du caractere recu
    case 'r' :                          // raz distance a la prochaine IT
      init_dist = true;                 
      break;
    case 'A' :                          // allumage feu arriere
    case 'a' :
      digitalWrite(lanterne_pin, HIGH); 
      break;
    case 'e' :                          // extinction feu arriere
      digitalWrite(lanterne_pin, LOW);
      break;
    case 'u' :                          // changement d'unite d'affichage.
                                        // les calculs sont toujours en mm et mm/s  
      switch (unit_mm++) {              // phases du cycle des unites d'affichage avec incrementation
        case 0:                         // on passe de mm/s a km/h on reste en m
          coef_vit = COEF_VIT;          // mise à jour du coef d'affichage de la vitesse
          vit[6]='k';                   // on remplace le m et le s de 'mm/s' pour obtenir 'km/h'
          vit[9]='h';
          break;
        case 1:                         // on passe de mm a m
          coef_dist=COEF_DIST;          // mise à jour du coef d'affichage de la distance
          dist[9]=' ';                  // on efface un m pour passer de 'mm' a 'm '
          break;
        case 2:                         // on revient aux unites de calcul
          coef_dist = 1.;               // les coefs reviennent à 1.
          coef_vit = 1.;
          dist[9]='m';                  // et les chaines de caracteres sont retablies            
          vit[6]='m';
          vit[9]='s';
          break;                
       }
      unit_mm%=3;                       // seulement 3 cas d'affichage (si unit_mm vaut 3 on revient a 0)
      //     init_dist = true;          // raz distance a la prochaine IT  
      break;   
  }   // fin du swich(c) 
}   // fin de serialEvent

void setup()  {
  Serial.begin(9600);               // initialize serial connection to PC
  while (!Serial) { ; }             // wait for serial port to connect. Needed for native USB port only
  
  // SSD1306_SWITCHCAPVCC = generate display voltage from 3.3V internally
  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) {   // Initialisation de l'objet afficheur a faire en premier car
                                                      // besoin de memoire contigue pour l'allocation dynamique
                                                      // du buffer d'affichage de 1024 octets
    Serial.println(F("SSD1306 allocation failed"));   // Init ratee
    for (;;);                                         // on s'arrete, boucle infinie
  }

  // affichages varies
  display.display();                    // affichage du logo de la librairie (splash.h) mis dans le buffer par l'init
  delay(2000);                          // pause de 2 seconde pour admirer le logo ;-)
  display.clearDisplay();               // effacement du buffer de l'afficheur
  display.setTextSize(2);               // taile 2 pour message lisible
  display.setTextWrap(true);            // autorise le retour à la ligne automatique
  display.setTextColor(WHITE);          // dessine le texte en blanc
  display.setCursor(0, 0);              // ecriture a partir du coin en haut à gauche
  display.println(F("Setup OK!" ));     // message de bienvenue sur l'ecran
  display.display();                    // affichage du buffer
  Serial.println(F("Setup OK" ));       // message de bienvenue sur la com serie
  delay(1000);                          // pause pour voir le message
  display.clearDisplay();               // effacement du buffer
  display.drawPixel(0, 0, WHITE);       // affiche un pixel en haut à gauche pour ne pas laisser l'ecran vide
  display.display();                    // affichage du buffer
  display.clearDisplay();               // efface le buffer pour le trace suivant
  display.setTextSize(1);               // remplit le buffer avec le titre de chaque valeur qui sera affichee
  display.setCursor(0, 0);              // en haut à gauche
  display.print(F("distance"));
  display.setCursor(0, 32);             // au milieu à gauche
  display.print(F("vitesse"));
  display.setTextSize(2);               // gros texte pour les valeurs mesurées

  aff_date = millis();                  // init date de reference pour les affichages OLED, PC

  // initialisation interruptions cadencees
  FlexiTimer2::set(2, 0.001, interruption);    // un tick toutes les 2*1ms pour executer interruption()
  FlexiTimer2::start();                 // on demarre les IT

  // initialisation des broches de sortie
  pinMode(led_pin, OUTPUT);             // led clignotante
  pinMode(lanterne_pin, OUTPUT);        // led feu arriere wagon
  pinMode(test_IT_pin, OUTPUT);         // sortie chronometrage IT
  pinMode(test_Form_pin, OUTPUT);       // sortie chronometrage Formatage
  pinMode(test_Aff_pin, OUTPUT);        // sortie chronometrage Affichage
  pinMode(test_Com_pin, OUTPUT);        // sortie chronometrage Communication serie
}


void loop() {  // la seule action de loop est de cadencer l'affichage des resultats
  if (millis() > aff_date)  {           // il est temps d'afficher
    aff_date += AFF_UPDATE;             // date de la prochaine action
    mise_en_forme();                    // mise en forme
    envoi();               				// envoi sur la liaison serie
    affichage();                        // affichage 
  }
  
}

void mise_en_forme() {// mise en forme de chaine de caracteres des resultats distance et vitesse dans dist et vit
  int i;                                // indice pour les boucles
  long data;                            // pour stocker la valeur a afficher
  char *p;                              // pointeur sur la chaine resultat du formatage
  char c;                               // caractere intermediaire (recu ou calcule pour le formatage)
digitalWrite(test_Form_pin, HIGH);      // mise a 1 de la sortie de chronometrage du formatage (au scope)
  // on commence par la distance
        // on remplit la chaine dist avec les 7 chiffres, sans toucher au nom de l'unite deja à la fin de dist ( voir setup() )
        data = long(distance*coef_dist + 0.5);    // la valeur de la distance est arrondie à l'unite
        if (data<0) {                   // si la valeur est négative :
          c='-';                        //      - on stocke le caractere du signe 
          data=-data;                   //      - on travaille sur la valeur absolue
        }
        else c=' ';                     // sinon, le caractere de signe est ' '
        p=dist;                         // p pointe sur le debut de la chaine resultat du format
        for (i=0;i<7;i++) *p++=' ';     // effacement de toute la chaine, a la fin p pointe sur la fin
        for (i=0;i<7;i++) {             // on remplit la chaine à partir de la fin (il y a toujours au moins 1 caractere)
          *p--=(data%10)+48;            // avec le reste de la division par 10 + 48 pour obtenir des caracteres
          data=data/10;                 // et on se prepare pour le chiffre suivant
          if (data==0) {                // il n'y a plus de chiffres a afficher
            *p=c;                       // on place le signe
            break;                      // et on sort de la boucle en laissant les espaces avant le signe
          }
        }
        if (data!=0) {                  // data etait superieur à 10^7 (il reste qquechose apres 7 divisions par 10)
          *p=dist;                      // il y a debordement :
          *p++=c;                       // on garde le signe
          *p++=' ';
          for (i=0;i<4;i++) *p++='X';   // et on met quelques X (4)
          *p++=' ';
          *p++=' ';
        }    
  // idem pour la vitesse (2 digits en moins)
        // on remplit la chaine vit avec les 5 chiffres, sans toucher au nom de l'unite deja à la fin de vit
        data = long(vitesse * coef_vit + 0.5);     // la valeur de la vitesse est arrondie à l'unite
        if (data<0) {                   // si la valeur est négative :
          c='-';                        //      - on prepare le caractere du signe 
          data=-data;                   //      - on travaille sur la valeur absolue
        }
        else c=' ';                     // sinon, le caractere de signe est ' '
        p=vit;                          // p pointe sur le debut de la chaine resultat du format
        for (i=0;i<5;i++) *p++=' ';     // effacement de toute la chaine, a la fin p pointe sur la fin de 
        for (i=0;i<5;i++) {             // on remplit la chaine à partir de la fin
          *p--=(data%10)+48;            // avec le reste de la division par 10 + 48 pour obtenir des caracteres
          data=data/10;                 // et on se prepare pour le chiffre suivant
          if (data==0) {                // il n'y a plus de chiffres a afficher
            *p=c;                       // on place le signe
            break;                      // et on sort de la boucle
          }
        }
        if (data!=0) {                  // data etait superieur à 10^5 (il reste qquechose apres 5 divisions par 10)
          *p=vit;                       // il y a debordement :
          *p++=c;                       // on garde le signe
          *p++=' ';
          for (i=0;i<4;i++) *p++='X';   // et on met quelques X
          *p++=' ';
        }
digitalWrite(test_Form_pin, LOW);   // mise à 0 de la sortie de chronometrage de la mise en forme
}

void affichage() {    // pilotage de l'afficheur AFF_OLED:

digitalWrite(test_Aff_pin, HIGH);   // mise à 1 de la sortie de chronometrage du pilotage de l'afficheur
        // toutes les actions se font dans le buffer avant de declencher l'affichage du buffer
        display.fillRect(0, 10, 128, 20,BLACK);   // on efface la valeur precedente dist par un rectangle noir 
        display.fillRect(0, 42, 128, 20,BLACK);   // on efface la valeur precedente vit par un rectangle noir
        display.setCursor(0, 10);       // on se positionne début 2eme ligne
        display.print(dist);            // on ecrit la valeur dans le buffer
        display.setCursor(0, 42);       // on se positionne début 4eme ligne
        display.print(vit);             // on ecrit la valeur dans le buffer
        display.display();              // affichage du buffer
digitalWrite(test_Aff_pin, LOW);   // mise à 0 de la sortie de chronometrage du pilotage de l'afficheur
}

void envoi() {                     // communication via la liaison serie
// COMM_Serie:                          
digitalWrite(test_Com_pin, HIGH);  // mise à 1 de la sortie de chronometrage du pilotage de la liaison serie
        Serial.write(dist,10);          // envoi des 10 caracteres de dist  (write() est plus rapide que print)
        Serial.write(",",1);            // envoi du separateur 
        Serial.write(vit,10);           // envoi des 10 caracteres de vit
        Serial.write(10);               // envoi de NewLine  \n
digitalWrite(test_Com_pin, LOW);  // mise à 0 de la sortie de chronometrage du pilotage de la liaison serie
}
